"""AttriBox provides a descriptor with lazy instantiation of the
underlying object. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static.zeroton import DELETED
from ..waitaminute import TypeException
from . import AbstractBox

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class AttriBox(AbstractBox):
  """AttriBox provides a descriptor with lazy instantiation of the
  underlying object. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  @classmethod
  def __class_getitem__(cls, wrapped: type) -> Self:
    """Get the field type."""
    self = cls.__new__(cls)
    self._setWrappedClass(wrapped)
    return self

  def __call__(self, *args, **kwargs) -> AttriBox:
    """Call the descriptor with the given arguments."""
    self.__init__(*args, **kwargs)
    return self

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_set__(self, val: Any, oldVal: Any = None, **kwargs) -> None:
    """
    Strongly typed setter requiring the new value to be a strict instance
    of the wrapped class.
    """
    if oldVal is None and not kwargs.get('_recursion', False):
      try:
        oldVal = self.__instance_get__()
      except Exception as exception:
        oldVal = exception
      else:
        pass
      finally:
        return self.__instance_set__(val, oldVal, _recursion=True, **kwargs)
    cls = self.getWrappedClass()
    if not isinstance(val, cls) and val is not DELETED:
      name = self.__field_name__
      raise TypeException(name, val, cls)
    pvtName = self.getPrivateName()
    setattr(self.instance, pvtName, val)

  def __instance_delete__(self, oldVal: Any = None, **kwargs) -> None:
    """
    Strongly typed deleter requiring the value to be deleted to be an
    instance of the wrapped class.
    """
    self.__instance_set__(DELETED, )



"""
_Priority exposes the priority of the accessor hook through the descriptor
protocol.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ...parse import maybe
from ...static import AbstractObject
from ...waitaminute import TypeException, VariableNotNone

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, TypeAlias
  from . import AbstractDescriptorHook as Hook
 
  Hooks: TypeAlias = tuple[Hook, ...]


class _Priority(AbstractObject):
  """
  _Priority exposes the priority of the accessor hook through the descriptor
  protocol.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback Variables
  __fallback_name__ = '__priority_value__'
  __fallback_value__ = 255

  #  Private Variables
  __private_name__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getPrivateName(self) -> str:
    """
    Reimplements the private name.
    """
    return maybe(self.__private_name__, self.__fallback_name__)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, **kwargs, ) -> Any:
    """
    Looks for '__priority_value__' in the instance and returns it.
    """
    pvtName = self.getPrivateName()
    try:
      out = getattr(self.instance, pvtName)
    except AttributeError:
      return self.__fallback_value__
    else:
      return maybe(out, self.__fallback_value__)
    finally:
      pass

  def __instance_set__(self, value: Any, **kwargs) -> None:
    """
    Sets the '__priority_value__' in the instance.
    """
    pvtName = self.getPrivateName()
    if value is None:
      raise VariableNotNone(pvtName, value)
    if not isinstance(value, int):
      raise TypeException(pvtName, value, int)
    if (255 - value) * (0 - value) < 0:
      raise NotImplementedError('TODO: IntervalError')
    setattr(self.instance, pvtName, value)



"""
CompoundHooks provides an array of 'AbstractDescriptorHook' objects owned
by 'AbstractDescriptor' objects.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from random import shuffle

from ...parse import maybe
from ...static import AbstractObject
from . import AbstractDescriptorHook as Hook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Iterator

  Hooks: TypeAlias = tuple[Hook, ...]
  HookList: TypeAlias = list[Hook]


class CompoundHooks(AbstractObject):
  """
  CompoundHooks provides an array of 'AbstractDescriptorHook' objects owned
  by 'AbstractDescriptor' objects.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback Variables

  #  Private Variables
  __accessor_hooks__ = None
  __iter_contents__ = None

  #  Public Variables

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getAccessorHooks(self) -> Hooks:
    """Get the accessor hooks."""
    return maybe(self.__accessor_hooks__, ())

  def _getOrderedHooks(self, ) -> HookList:
    """
    Get the ordered hooks.

    This method returns the hooks sorted by priority in descending order.
    """
    hooks = self._getAccessorHooks()
    return [*sorted(hooks, key=lambda h: h.priority, ), ]

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addAccessorHook(self, hook: Hook) -> Self:
    """Add an accessor hook."""
    existing = self.__accessor_hooks__
    self.__accessor_hooks__ = (*existing, hook)
    return self

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __iter__(self) -> Self:
    """
    Yields hooks in order of priority with hooks of same priority randomly
    chosen.
    """
    hooks = self._getAccessorHooks()
    priorityMap = dict()
    for hook in hooks:
      existing = priorityMap.get(hook.priority, [])
      existing.append(hook)
      priorityMap[hook.priority] = existing
    self.__iter_contents__ = []
    priorities = sorted(priorityMap.keys(), reverse=True)
    for p in priorities:
      self.__iter_contents__.extend(priorityMap[p])
    return self

  def __next__(self, ) -> Hook:
    """
    Returns the next hook in the iteration. If there are no more hooks,
    raises StopIteration.
    """
    if self.__iter_contents__:
      return self.__iter_contents__.pop(0)
    self.__iter_contents__ = None
    raise StopIteration



"""
TypedAttr subclasses Attr adding strong types.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..parse import maybe
from ..waitaminute import MissingVariable, TypeException
from . import Attr

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class TypedAttr(Attr):
  """Creates a deferred function that is called when the __get__ is first
  called. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __abstract_type__ = object

  #  Private variables
  __field_type__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getFieldType(self, **kwargs) -> type:
    """
    Getter-function for the field type.
    """
    return maybe(self.__field_type__, self.__abstract_type__, )

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _setFieldType(self, fieldType: type) -> Self:
    """
    Setter-function for the field type.
    """
    if self.__field_type__ is not None:
      raise MissingVariable('__field_type__', type)
    if not isinstance(fieldType, type):
      name, value = '__field_type__', fieldType
      raise TypeException(name, value, type)
    self.__field_type__ = fieldType
    return self

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, ) -> None:
    name, type_ = None, None
    for arg in args:
      if isinstance(arg, str) and name is None:
        name = arg
        continue
      elif isinstance(arg, type) and type_ is None:
        type_ = arg
        continue
      if all([i is not None for i in (name, type_)]):
        break
    else:
      if name is None:
        name = self._getPrivateName()
      type_ = maybe(type_, self.__abstract_type__, )
    Attr.__init__(self, name)
    self._setFieldType(type_)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _typeGuard(self, value: Any) -> Any:
    """
    Type guard for the value.
    """
    fieldType = self._getFieldType()
    if not isinstance(value, fieldType):
      name, value = '_typeGuard', value
      raise TypeException(name, value, fieldType)
    return value

  def __instance_get__(self, **kwargs) -> Any:
    """
    Getter-function for the instance.
    """
    return self._typeGuard(super().__instance_get__(**kwargs))

  def __instance_set__(self, value: Any, **kwargs) -> Self:
    """
    Setter-function for the instance.
    """
    return super().__instance_set__(self._typeGuard(value), **kwargs)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""NumHook subclasses AbstractHook and provides a hook for instances of
NUM. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import typeMsg
from ..waitaminute import ReservedName
from ..mcls import FunctionType as Func
from ..mcls.hooks import AbstractHook
from . import NUM

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any
  from worktoy.keenum import NumSpace as NSpace


class NumHook(AbstractHook):
  """NumHook subclasses AbstractHook and provides a hook for instances of
  NUM. """

  @staticmethod
  def _getReservedNames() -> list[str]:
    """Get the reserved names for the enumeration."""
    return [
        '__init__',
        '__bool__',
        '__str__',
        '__repr__',
        '__eq__',
        '__get__',
        '__set__',
        '__delete__',
    ]

  def setItemHook(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """The setItemHook method is called when an item is set in the
    enumeration."""
    if key in self._getReservedNames():
      if not callable(value):
        raise TypeError(typeMsg(key, value, Func))
      if getattr(value, '__is_root__', None) is None:
        raise ReservedName(key, )
      if self.space.getClassName() != 'KeeNum':
        raise ReservedName(key, )
      return False
    if isinstance(value, NUM):
      return self.space.addNum(key, value)
    return False

  def postCompileHook(self, compiledSpace: dict) -> dict:
    """The postCompileHook method is called after the namespace is
    compiled."""
    compiledSpace['__member_objects__'] = []
    compiledSpace['__allow_instantiation__'] = True
    return compiledSpace



"""
CallMeMaybe is like typing.Callable except usable at runtime.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class CallMeMaybe:
  """CallMeMaybe is like typing.Callable except usable at runtime.

  It is used to represent a callable object that can be called with
  any number of arguments and keyword arguments.
  """

  def __instancecheck__(self, instance: Any) -> bool:
    """Check if the instance is callable."""
    return True if callable(instance) else False



"""ResolveException should be raised by classes that try to resolve
'other' objects in custom implementation of certain dunder methods. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any


class ResolveException(Exception):
  """ResolveException should be raised by classes that try to resolve
  'other' objects in custom implementation of certain dunder methods. """

  selfCls = _Attribute()
  otherObj = _Attribute()

  def __init__(self, self_: Any, other: Any) -> None:
    """Initialize the ResolveException object."""
    self.selfCls = type(self_)
    self.otherObj = other
    clsName = self.selfCls.__name__
    otherStr = repr(other)
    info = """Unable to resolve '%s' as an object of type: '%s'"""
    Exception.__init__(self, info % (otherStr, clsName))



"""THISContextException is a custom exception raised to indicate a disallowed
use of the 'THIS' token object. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import joinWords, monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Callable

  from ..attr import _FieldProperties as Desc


class THISContextException(ValueError):
  """THISContextException is a custom exception raised to indicate a
  disallowed use of the 'THIS' token object. """

  context = _Attribute()

  def __init__(self, context: Callable) -> None:
    """Initialize the THISContextException object."""
    self.context = context
    infoSpec = """In the context of '%s', the THIS token object is not 
    supported!"""
    if hasattr(context, '__qualname__'):
      info = monoSpace(infoSpec % (context.__qualname__,))
    else:
      info = monoSpace(infoSpec % (context.__name__,))
    ValueError.__init__(self, info)

  def __eq__(self, other: Any) -> bool:
    """Compare the THISContextException object with another object."""
    cls = type(self)
    if not isinstance(other, cls):
      return False
    if other is self:
      return True
    if other.context != self.context:
      return False
    return True



"""
The 'worktoy.attr.field' module provides the classes used by the
comprehensive 'Field' descriptor class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations



"""AbstractNamespace class provides a base class for custom namespace
objects used in custom metaclasses."""
#  AGPL-3.0 license
#  Copyright (c) 2024-2025 Asger Jon Vistisen
from __future__ import annotations

from ..parse import maybe
from ..static import HistDict
from ..text import typeMsg, monoSpace, joinWords
from ..waitaminute import MissingVariable, HashError, TypeException
from . import Base, Types, Spaces
from .hooks import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class AbstractNamespace(HistDict):
  """AbstractNamespace class provides a base class for custom namespace
  objects used in custom metaclasses."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __owner_hooks_list_name__ = '__hook_objects__'

  #  Reserved private names
  __meta_class__ = None
  __class_name__ = None
  __base_classes__ = None
  __base_spaces__ = None
  __key_args__ = None
  __hash_value__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  @classmethod
  def getHookListName(cls, ) -> str:
    """Getter-function for the name of the hook list. """
    if TYPE_CHECKING:
      assert isinstance(cls, dict)
    return cls.__owner_hooks_list_name__

  def getHooks(self, owner: type = None) -> list[AbstractHook]:
    """Getter-function for the AbstractHook classes. """
    pvtName = self.getHookListName()
    cls = type(self)
    hooks = getattr(cls, pvtName, [])
    out = []
    for hook in hooks:
      out.append(hook.__get__(self, cls))
    return out

  @classmethod
  def getLegacySpace(cls, self, ) -> Self:
    """
    Returns the namespace object of the base classes.
    """
    itemCalls = []
    mroClasses = [*reversed(self.getMRO(), )]
    for base in mroClasses:
      if base is self:
        continue
      space = getattr(base, '__namespace__', None)
      if space is None:
        continue
      if isinstance(space, HistDict):
        itemCalls.extend(space.getItemCalls())
        continue
    mcls = self.getMetaclass()
    name = self.getClassName()
    bases = self.getClassBases()
    kwargs = self.getKwargs()
    out = cls(mcls, name, (), **kwargs)
    for call in itemCalls:
      out = call.applyToNamespace(out)
    return out

  def getClassBases(self) -> tuple[type, ...]:
    """
    Getter-function for the given base classes.
    """
    return maybe(self.__base_classes__, ())

  def _computeHash(self, ) -> int:
    """
    Computes the hash value for the future class.
    """
    clsName = self.getClassName()
    baseNames = [base.__name__ for base in self.getClassBases()]
    mclsName = self.getMetaclass().__name__
    return hash((clsName, *baseNames, mclsName,))

  def getHash(self, **kwargs) -> int:
    """
    Resolves the hash value for the future class.
    """
    computedHash = self._computeHash()
    if self.__hash_value__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.__hash_value__ = computedHash
      return self.getHash(_recursion=True)
    if not isinstance(self.__hash_value__, int):
      raise TypeException('__hash_value__', self.__hash_value__, int)
    if self.__hash_value__ != computedHash:
      raise HashError(self.__hash_value__, computedHash, )
    return self.__hash_value__

  def getMetaclass(self, ) -> type:
    """Returns the metaclass."""
    return self.__meta_class__

  def getClassName(self, ) -> str:
    """Returns the name of the class."""
    return self.__class_name__

  def getKwargs(self, ) -> dict:
    """Returns the keyword arguments passed to the class."""
    return {**self.__key_args__, **dict()}

  def getMRO(self, ) -> Types:
    """Returns the method resolution order of the class."""
    mcls = self.getMetaclass()
    name = self.getClassName()
    Space = type(self)
    space = Space(mcls, name, self.getClassBases(), **self.getKwargs())
    bases = self.getClassBases()
    tempClass = mcls('_', (*bases,), space)
    return tempClass.__mro__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def addHook(cls, hook: AbstractHook) -> None:
    """Adds a hook to the list of hooks. """
    if TYPE_CHECKING:
      assert isinstance(cls, dict)
    if not isinstance(hook, AbstractHook):
      raise TypeError(typeMsg('hook', hook, AbstractHook))
    existingHooks = getattr(cls, cls.getHookListName(), [])
    setattr(cls, cls.getHookListName(), [*existingHooks, hook])

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, mcls: type, name: str, bases: Base, **kwargs) -> None:
    self.__meta_class__ = mcls
    self.__class_name__ = name
    if len(bases) > 1:
      e = """The 'worktoy' package does not support multiple inheritance!"""
      raise TypeError(monoSpace(e))
    self.__base_classes__ = (*bases,)
    self.__key_args__ = kwargs or {}

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __setitem__(self, key: str, val: object, **kwargs) -> None:
    """Sets the value of the key."""
    try:
      oldVal = dict.__getitem__(self, key)
    except KeyError:
      oldVal = None
    for hook in self.getHooks():
      if TYPE_CHECKING:
        assert isinstance(hook, AbstractHook)
      if hook.setItemHook(self, key, val, oldVal):
        break
    else:
      dict.__setitem__(self, key, val)

  def _getStrComponents(self, ) -> dict:
    """
    Resolves the components of 'self' that is shared between __repr__ and
    __str__.
    """
    spaceClass = type(self).__name__
    className = self.getClassName()
    mclsName = self.getMetaclass().__name__
    baseNames = [base.__name__ for base in self.getClassBases()]

    kwargStr = []
    for key, value in self.getKwargs().items():
      kwargStr.append("""%s=%s""" % (key, repr(value),))
    kwargStr = ', '.join(kwargStr, )

    return dict(
        spaceClass=spaceClass,
        className=className,
        mclsName=mclsName,
        baseName=baseNames,
        kwargStr=kwargStr,
    )

  def __str__(self, ) -> str:
    """Returns the string representation of the namespace object."""
    infoSpec = """Namespace object of type: '%s' created by metaclass: 
    '%s' with bases: '%s' to create class: '%s'."""
    components = self._getStrComponents()
    spaceClass = components['spaceClass']
    className = components['className']
    mclsName = components['mclsName']
    baseName = joinWords(components['baseName'], )
    kwargStr = components['kwargStr']
    return infoSpec % (spaceClass, mclsName, baseName, className)

  def __repr__(self, ) -> str:
    """Returns the string representation of the namespace object."""
    infoSpec = """%s(%s, %s, %s, **%s)"""
    components = self._getStrComponents()
    spaceClass = components['spaceClass']
    className = components['className']
    mclsName = components['mclsName']
    baseName = joinWords(components['baseName'], )
    kwargStr = components['kwargStr']
    return infoSpec % (spaceClass, mclsName, className, baseName, kwargStr,)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def preCompile(self, ) -> dict:
    """The return value from this method is passed to the compile method.
    Subclasses can implement this method to provide special objects at
    particular names in the namespace. By default, an empty dictionary is
    returned. """
    namespace = dict()
    for hook in self.getHooks():
      if TYPE_CHECKING:
        assert isinstance(hook, AbstractHook)
      namespace = hook.preCompileHook(self, namespace)
    return namespace

  def compile(self, ) -> dict:
    """This method is responsible for building the final namespace object.
    Subclasses may reimplement preCompile or postCompile as needed,
    but must not reimplement this method."""
    namespace = self.preCompile()
    for (key, val) in dict.items(self, ):
      namespace[key] = val
    namespace['__metaclass__'] = self.getMetaclass()
    namespace['__namespace__'] = self
    return self.postCompile(namespace)

  def postCompile(self, namespace: dict) -> dict:
    """The object returned from this method is passed to the __new__
    method in the owning metaclass. By default, this method returns dict
    object created by the compile method after performing certain
    validations. Subclasses can implement this method to provide further
    processing of the compiled object. """
    for hook in self.getHooks():
      if TYPE_CHECKING:
        assert isinstance(hook, AbstractHook)
      namespace = hook.postCompileHook(self, namespace)
    if '__metaclass__' not in namespace:
      raise MissingVariable('__metaclass__', self.getMetaclass())
    if '__namespace__' not in namespace:
      raise MissingVariable('__namespace__', type(self))
    return namespace



"""WaitForIt provides a deferred descriptor.
Usage:
class Foo:
  bar = WaitForIt(func, *args, **kwargs)
Then:
foo = Foo()
foo.bar = Foo.bar.__get__(foo, Foo) = func(*args, **kwargs)

This allows lazy evaluation of the given function with the given
arguments. By passing 'THIS' as an argument, 'THIS' is replaced with the
instance passed to __get__.

By passing only a 'str' object as the first argument, __get__ returns:
getattr(instance, arg: str).

The first argument must be a callable, a str object pointing to a key or
THIS, which is redundant.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import MissingVariable, TypeException, VariableNotNone
from . import Field

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any
  from typing import Callable as Func
else:
  Func = type('_', (type,), dict(__instancecheck__=callable))('_', (), {})


class WaitForIt(Field):
  """Creates a deferred function that is called when the __get__ is first
  called. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __creator_function__ = None  # The function that creates the descriptor

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getCreatorFunction(self) -> Func:
    """Get the creator function."""
    if self.__creator_function__ is None:
      raise MissingVariable('__creator_function__', Func)
    if not isinstance(self.__creator_function__, Func):
      name, value = '__creator_function__', self.__creator_function__
      raise TypeException(name, value, Func)
    #  _getCreatorFunction should never be invoked, except when both owner
    #  and instance have objects other than None. Thus, the creator may be
    #  retrieved from the instance ensuring it is a bound method.
    #  Therefore, the 'self/cls' first argument will not swallow one of
    #  the expected values. If the function is not present, it means it is
    #  not a method in the class. In this case, the function is called
    #  directly.
    key = self.__creator_function__.__name__
    return getattr(self.instance, key, self.__creator_function__)

  def _hasCreatorFunction(self, ) -> bool:
    """Check if the creator function is set."""
    if self.__creator_function__ is None:
      return False
    if isinstance(self.__creator_function__, Func):
      return True
    name, value = '__creator_function__', self.__creator_function__
    raise TypeException(name, value, Func)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _setCreatorFunction(self, creator: Func) -> None:
    """Set the creator function."""
    if self.__creator_function__ is not None:
      name, value = '__creator_function__', self.__creator_function__
      raise VariableNotNone(name, value)
    if not isinstance(creator, Func):
      name, value = '__creator_function__', creator
      raise TypeException(name, value, Func)
    self.__creator_function__ = creator

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, **kwargs) -> None:
    posArgs = []
    tempArgs = [*args, ]
    while tempArgs:
      arg = tempArgs.pop(0)
      if isinstance(arg, Func):
        self._setCreatorFunction(arg)
        posArgs.extend(tempArgs)
        Field.__init__(self, *posArgs, **kwargs)
        break
      posArgs.append(arg)
    else:
      Field.__init__(self, *args, **kwargs)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, **kwargs) -> Any:
    """
    Get the value of the descriptor, which is the result of calling the
    creator function with the given arguments.
    """
    creator = self._getCreatorFunction()
    pvtName = self.getPrivateName()
    if not hasattr(self.instance, pvtName, ):
      if kwargs.get('_recursion', False):
        raise RecursionError
      args = self._getPositionalArgs()
      kwargs = self._getKeywordArgs()
      value = creator(*args, **kwargs)
      setattr(self.instance, pvtName, value)
      return self.__instance_get__(_recursion=True, )
    return getattr(self.instance, pvtName, )

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __call__(self, func: Func) -> Any:
    """Call the descriptor with the given arguments."""
    self._setCreatorFunction(func)
    return func



"""
PreClass provides a stateful class containing the name and hash of a class
about to be created.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..parse import maybe
from ..waitaminute import MissingVariable, VariableNotNone, TypeException

from . import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class PreClass(AbstractObject):
  """PreClass provides a stateful class containing the name and hash of a
  class about to be created."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __private_fallback__ = '__pre_class__'

  #  Private variables
  __private_name__ = None
  __class_name__ = None
  __hash_value__ = None

  #  Public variables

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, **kwargs) -> None:
    AbstractObject.__init__(self, *args, **kwargs)
    _hash, _name, _meta = None, None, None
    for arg in args:
      if isinstance(arg, str):
        if _name is None:
          _name = arg
          continue
        raise VariableNotNone('_name')
      if isinstance(arg, int):
        if _hash is None:
          _hash = arg
          continue
        raise VariableNotNone('_hash')
      if isinstance(arg, type):
        if _meta is None:
          _meta = arg
          continue
        raise VariableNotNone('_meta')
      if all([i is not None for i in (_name, _hash, _meta)]):
        break
    else:
      if _name is None:
        raise MissingVariable('_name', str)
      if _hash is None:
        raise MissingVariable('_hash', int)
      if _meta is None:
        raise MissingVariable('_meta', type)
    self.__hash_value__ = _hash
    self.__class_name__ = _name
    self.__meta_class__ = _meta

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __hash__(self, ) -> int:
    """
    Returns the explicitly set hash value of the PreClass object.
    """
    if self.__hash_value__ is None:
      raise MissingVariable('__hash_value__', int)
    if isinstance(self.__hash_value__, int):
      return self.__hash_value__
    name, value = '__hash_value__', self.__hash_value__
    raise TypeException(name, value, int, )

  def __getattribute__(self, key: str, ) -> Any:
    """
    This reimplementation of __getattribute__ was done by a highly skilled
    professional, do not try this at home!
    """
    if key == '__name__':
      return object.__getattribute__(self, '__class_name__')
    if key == '__class__':
      return object.__getattribute__(self, '__meta_class__')
    return AbstractObject.__getattribute__(self, key)



"""
AbstractBox provides a descriptor with lazy instantiation of a particular
class instantiating only when 'AbstractBox.__get__' receives an instance
(not 'None'). This conserves resources and streamlines object creation.

Qt applications such as developed by Pyside6, require the
'QCoreApplication' to be running before any 'QObject' may be instantiated.
Since Pyside6 is effectively a C++ library, failure to adhere to this
requirement leads to segmentation related errors occurring separate from
the Python interpreter. In other words, it leads to highly undefined
behaviour. By using 'AbstractBox' descriptors when developing in Pyside6,
this design patterns becomes much easier to implement and maintain as the
descriptor itself defers instantiation.

AbstractBox descriptor implements only 'get' functionality, which
looks for an instance of the wrapped class at the private name of
the descriptor in the namespace of the current instance. If no such is
found, it creates one and recursively calls the instance get again. This
attempt passes positional and keyword arguments received by the descriptor
constructor and passes them on to the wrapped class constructor.

The 'set' functionality may be implemented in multiple different ways,
unlike the 'get' functionality which AbstractBox does provide. For 'set'
and 'delete' functionality, AbstractBox leaves it to subclasses to change
the default behaviour of simply raising 'ReadOnlyError' or
'ProtectedError' respectively.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static.zeroton import DELETED
from ..waitaminute import MissingVariable, TypeException, \
  attributeErrorFactory
from . import Field

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, TypeAlias

  Types: TypeAlias = tuple[type, ...]


class AbstractBox(Field):
  """AttriBox provides a descriptor with lazy instantiation of the
  underlying object. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __wrapped__ = None  # __get__ returns an instance of this class

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getWrappedClass(self) -> type:
    """
    Getter-function for the wrapped class.
    """
    if self.__wrapped__ is None:
      raise MissingVariable('__wrapped__', type)
    if isinstance(self.__wrapped__, type):
      return self.__wrapped__
    name, value = '__wrapped__', self.__wrapped__
    raise TypeException(name, value, type)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _setWrappedClass(self, wrapped: type) -> None:
    """
    Setter-function for the wrapped class.
    """
    if self.__wrapped__ is not None:
      raise MissingVariable('__wrapped__', type)
    if not isinstance(wrapped, type):
      name, value = '__wrapped__', wrapped
      raise TypeException(name, value, type)
    self.__wrapped__ = wrapped
    return

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _createdWrappedObject(self, ) -> Any:  # of __wrapped__ type
    """
    Creator-function for the object of __wrapped__ type.
    """
    args = self._getPositionalArgs()
    kwargs = self._getKeywordArgs()
    wrapCls = self.getWrappedClass()
    return wrapCls(*args, **kwargs)

  def __instance_get__(self, **kwargs) -> Any:
    """
    Instance getter-function retrieves a wrapped object already at the
    current instance. If no wrapped object is found, an attempt is made to
    create and set one after which __'instance_get__' is called again,
    but with the '_recursion' keyword argument set to 'True'. If again, no
    wrapped object is found, the function raises a plain RecursionError.
    """
    pvtName = self.getPrivateName()
    out = getattr(self.instance, pvtName, None)
    if out is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      out = self._createdWrappedObject()
      setattr(self.instance, pvtName, out)
      return self.__instance_get__(_recursion=True, )
    if out is DELETED:
      raise attributeErrorFactory(self.owner, self.__field_name__)

    wrapped = self.getWrappedClass()
    if isinstance(out, wrapped):
      return out
    name, value = pvtName, out
    raise TypeException(name, value, wrapped)



"""
FlexBox provides a dynamically typed version of AttriBox.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import AttriBox
from ..waitaminute import TypeException, CascadeException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class FlexBox(AttriBox):
  """
  FlexBox provides a dynamically typed version of AttriBox.
  """

  def __instance_set__(self, val: Any, oldVal: Any = None, **kwargs) -> None:
    """
    Strongly typed setter requiring the new value to be a strict instance
    of the wrapped class.
    """
    try:
      AttriBox.__instance_set__(self, val, oldVal, **kwargs)
    except TypeException as typeException:
      cls = self.getWrappedClass()
      if typeException.varName == self.__field_name__:
        if typeException.actualObject == val:
          cls = self.getWrappedClass()
          exp = typeException.expectedType
          if cls in exp:
            try:
              value = cls(val)
            except Exception as exception:
              try:
                value = cls(*val)
              except Exception as exception2:
                errors = [
                    typeException,
                    exception,
                    exception2
                ]
                raise CascadeException(*errors)
            else:
              self.__instance_set__(value, oldVal, _recursion=True)
          else:
            print(typeException.expectedType, cls)
      else:
        raise typeException

    else:
      return



"""
_KeeMember encapsulates future enumeration member objects. The 'auto'
function instantiates this class during the class body execution. The
control flow collects these objects during compilation of the final
namespace object. For this reason, _KeeMember is private to the
'worktoy.keenum' module. Instances of _KeeMember are discarded during the
class creation process.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..attr import Field
from ..mcls import AbstractNamespace, BaseObject
from ..parse import maybe
from ..static import AbstractObject, overload
from ..static.zeroton import THIS
from ..waitaminute import MissingVariable, TypeException, VariableNotNone, \
  IllegalInstantiation

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Any


class _KeeMember(BaseObject):
  """
  _KeeMember encapsulates future enumeration member objects. The 'auto'
  function instantiates this class during the class body execution. The
  control flow collects these objects during compilation of the final
  namespace object. For this reason, _KeeMember is private to the
  'worktoy.keenum' module. Instances of _KeeMember are discarded during the
  class creation process.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __member_name__ = None  # Future name of the enumeration member
  __member_value__ = None  # Future value of the enumeration member

  #  Public Variables
  name = Field()
  value = Field()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @name.GET
  def _getName(self) -> str:
    """Get the name of the enumeration member."""
    if self.__member_name__ is None:
      raise MissingVariable('__member_name__', str)
    if isinstance(self.__member_name__, str):
      return self.__member_name__
    raise TypeException('__member_name__', self.__member_name__, str)

  @value.GET
  def _getValue(self) -> Any:
    """Get the value of the enumeration member."""
    if self.__member_value__ is None:
      return self.__member_name__
    return self.__member_value__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @name.SET
  def _setName(self, name: str) -> None:
    """Set the name of the enumeration member."""
    if self.__member_name__ is not None:
      raise VariableNotNone('__member_name__', )
    if not isinstance(name, str):
      raise TypeException('__member_name__', name, str)
    self.__member_name__ = name

  @value.SET
  def _setValue(self, value: Any) -> None:
    """Set the value of the enumeration member."""
    if self.__member_value__ is not None:
      raise VariableNotNone('__member_value__', )
    self.__member_value__ = value

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, **kwargs) -> None:
    """
    The control flow instantiates and initializes the members
    automatically. Do not reimplement or change this function. Developers
    of 'worktoy' considers this function guaranteed to remain unchanged.
    Changing it causes undefined behavior. For this reason, future
    versions are planned to specifically raise RuntimeError if this
    function is reimplemented or changed.

    As of development version 0.99.85, this function could be
    reimplemented by setting the '_root' keyword argument to True,
    but a much better method of preventing instantiation is in the works.
    This new prevention scheme will be backwards compatible, ensuring that
    libraries depending on 'worktoy.keenum' can safely upgrade, provided
    they do not reimplement this function!
    """
    if not kwargs.get('_root', False):
      raise IllegalInstantiation(type(self))



"""
DuplicateSeedException is a custom exception raised by the KingDice class
when receiving duplicate seeds.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import monoSpace

try:
  from typing import TYPE_CHECKING, Any
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.worktest import KingDice


class DuplicateSeedException(Exception):
  """
  SeedException is a custom exception raised by the KingDice class when
  receiving duplicate seeds.
  """

  king = _Attribute()
  oldSeed = _Attribute()
  newSeed = _Attribute()

  def __init__(self, king: KingDice, oldSeed: int, newSeed: int) -> None:
    """
    Initializes the DuplicateSeedException with the KingDice instance and
    the old and new seeds.

    Args:
        king (KingDice): The KingDice instance that raised the exception.
        oldSeed (int): The old seed that caused the exception.
        newSeed (int): The new seed that caused the exception.
    """
    self.king = king
    self.oldSeed = oldSeed
    self.newSeed = newSeed
    fieldName = getattr(self.king, '__field_name__')
    fieldOwner = getattr(self.king, '__field_owner__')
    if fieldName is None or fieldOwner is None:
      infoSpec = """Tried changing the seed of existing '%s%s%s' object from 
      %d to %d. '%s' objects are expected to have unchanging seeds.  """
      ownerName = ''
      fieldName = ''
    else:
      infoSpec = """Tried changing the seed of existing '%s' object at 
      '%s.%s' from %d to %d. '%s' objects are expected to have unchanging 
      seeds. """
      ownerName = fieldOwner.__name__
    clsName = type(self.king).__name__
    info = infoSpec % (
        clsName, ownerName, fieldName, oldSeed, newSeed, clsName
    )
    Exception.__init__(self, monoSpace(info))

  def __eq__(self, other: Any) -> bool:
    """
    Checks if the current instance is equal to another object.

    Args:
        other (Any): The object to compare with.

    Returns:
        bool: True if the objects are equal, False otherwise.
    """
    if not isinstance(other, DuplicateSeedException):
      return False
    if self.king != other.king:
      return False
    if self.oldSeed != other.oldSeed:
      return False
    if self.newSeed != other.newSeed:
      return False
    return True



"""
_CallMeMaybe encapsulates function decoration.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from ..parse import maybe
from ..static import AbstractObject
from ..waitaminute import MissingVariable, TypeException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any


class _CallMeMaybe(AbstractObject):
  """_CallMeMaybe encapsulates function decoration."""

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __fallback_wrapper__ = lambda hereIsMyNumber: hereIsMyNumber

  #  Private variables
  __wrapper__ = None  # The function that will actually decorate
  __wrapped__ = None  # The function that is wrapped by the decorator

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getWrapper(self, ) -> Func:
    """
    Getter-function for the wrapper function.
    """
    return maybe(self.__wrapper__, self.__fallback_wrapper__)

  def _getWrapped(self) -> Func:
    """
    Getter-function for the wrapped function.
    """
    if self.__wrapped__ is None:
      raise MissingVariable('wrapped', Func)
    return self.__wrapped__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, wrapperFunction: Func) -> None:
    """Initialize the CallMeMaybe object."""
    if not isinstance(wrapperFunction, Func):
      raise TypeException('wrapper', wrapperFunction, Func)
    self.__wrapper__ = wrapperFunction

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __call__(self, *args, **kwargs) -> Any:
    """Call the wrapped function with the given arguments."""
    try:
      wrapped = self._getWrapped()
    except MissingVariable as mv:
      if len(args) == 1 and not kwargs:
        if isinstance(args[0], Func):
          self.__wrapped__ = args[0]
          return self
      raise mv
    else:
      return wrapped(*args, **kwargs)
    finally:
      pass



"""
AbstractDescriptor provides the base class for descriptors in the
'worktoy.attr' module.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from random import shuffle

from ..parse import maybe
from ..static import AbstractObject
from ..waitaminute import MissingVariable

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, TypeAlias, Any
  from .accessor_hooks import AbstractDescriptorHook as Hook

  Hooks: TypeAlias = dict[str, Hook]
  HookList: TypeAlias = list[Hook]


class AbstractDescriptor(AbstractObject):
  """
  'AbstractDescriptor' inherits descriptor protocol implementation from
  'AbstractObject' and enhances it with a hook system surrounding each
  accessor method. These hooks are class specific and thus set with
  classmethod 'addAccessorHook'. The intended use is for hook classes to
  register themselves during '__set_name__'.

  The control flow dispatches hooks in order determined by their priority
  attribute from low to high. Tie-breakers are intentionally given fuzzy
  ordering to avoid undefined deterministic behaviour concealing errors.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback Variables

  #  Private Variables
  __accessor_hooks__ = None

  #  Public Variables

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def _getAccessorHooks(cls) -> Hooks:
    """Get the accessor hooks."""
    return maybe(cls.__accessor_hooks__, dict())

  @classmethod
  def getOrderedHooks(cls, ) -> HookList:
    """
    Returns the hooks in order of descending priority. Hooks with the same
    priority are returned in random order (via shuffle). Lower numeric
    priority values are considered more urgent.
    """
    accessorHooks = cls._getAccessorHooks()
    priorityMap = dict()
    priorities = [h.priority for n, h in accessorHooks.items()]
    priorities = sorted(list(set(priorities)), reverse=True)
    out = []
    for name, hook in accessorHooks.items():
      if hook.priority not in priorityMap:
        priorityMap[hook.priority] = [hook, ]
        continue
      priorityMap[hook.priority].append(hook)
    for p, hooks in priorityMap.items():
      shuffle(hooks)
      out.extend(hooks)
    return out

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  @classmethod
  def registerAccessorHook(cls, name: str, hook: Hook) -> Self:
    """
    Adds the given hook at the given name. This is meant to be invoked by
    '__set_name__' on the hook class.
    """
    hooks = cls._getAccessorHooks()
    if name in hooks:
      raise NotImplementedError('Duplicate hook names lmao')
    hooks[name] = hook
    cls.__accessor_hooks__ = hooks
    return cls

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __get__(self, instance: Any, owner: Any, **kwargs) -> Any:
    """
    Reimplementation preserving the 'updateContext' call and return of
    'self' when accessed on class (instance is None). Invokes hooks
    before and after '__instance_get__' method.
    """
    self.updateContext(instance, owner, **kwargs)
    if instance is None:
      return self
    hooks = self.getOrderedHooks()
    #  Pre hooks
    preValue = None
    for hook in hooks:
      val = hook.__pre_get__()
      if val is not None:
        if preValue is None:
          preValue = val
          continue
        raise NotImplementedError('TODO: Multiple pre-get values')
    #  Collect value from pre hooks or from instance
    if preValue is None:
      value = self.__instance_get__(**kwargs)
    else:
      value = preValue
    #  Post hooks
    for hook in hooks:
      postValue = hook.__post_get__(value)
      if postValue is not None:
        return postValue
    return value

  def __set__(self, instance: Any, value: Any, **kwargs) -> None:
    """
    Reimplementation preserving the 'updateContext' call and invoking
    hooks before and after '__instance_set__' method.
    """
    self.updateContext(instance, **kwargs)
    hooks = self.getOrderedHooks()
    #  Pre hooks
    preValue = None
    for hook in hooks:
      val = hook.__pre_set__(value, )
      if val is not None:
        if preValue is None:
          preValue = val
          continue
        raise NotImplementedError('TODO: Multiple pre-set values')
    if preValue is not None:
      return self.__instance_set__(preValue, **kwargs)
    #  Collect old value or MissingValue exception
    try:
      oldVal = self.__instance_get__()
    except AttributeError as missingValue:
      oldVal = missingValue
    #  Return 'None' if value is identical to old value
    if value is oldVal:
      return
    #  Post hooks
    postValue = None
    for hook in hooks:
      val = hook.__post_set__(value, oldVal)
      if val is not None:
        if postValue is not None:
          raise NotImplementedError('TODO: Multiple post-set values')
        postValue = val
    if postValue is None:
      return self.__instance_set__(value, **kwargs)
    return self.__instance_set__(postValue, **kwargs)

  def __delete__(self, instance: Any, **kwargs) -> None:
    """
    Reimplementation preserving the 'updateContext' call and invoking
    hooks before '__instance_delete__' method.
    """
    self.updateContext(instance, **kwargs)
    hooks = self.getOrderedHooks()
    for hook in hooks:
      hook.__pre_delete__()
    try:
      oldValue = self.__instance_get__(**kwargs)
    except MissingVariable as missingVariable:
      raise AttributeError from missingVariable
    for hook in hooks:
      hook.__post_delete__(oldValue)
    return self.__instance_delete__(**kwargs)



"""
GET provides the decorator used by 'worktoy.attr.field.Field' to designate
the getter method in the class body.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from ...parse import maybe
from ...static import AbstractObject
from ...static.zeroton import THIS, OWNER, DESC
from ...waitaminute import MissingVariable, TypeException
from ...waitaminute import SubclassException
from ...waitaminute import ReadOnlyError, ProtectedError

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Union, Self, TypeAlias, Never


class GET(AbstractObject):
  """Creates a deferred function that is called when the __get__ is first
  called. """
 


"""
KeeHook provides the namespace hook for the KeeSpace namespace class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from ..parse import maybe
from ..static import AbstractObject
from ..static import AbstractObject as Object
from ..text import typeMsg
from ..waitaminute import ReservedName
from ..mcls.hooks import AbstractHook

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, TypeAlias

  from . import KeeSpace, _AutoMember


class KeeHook(AbstractHook):
  """
  KeeHook provides the namespace hook for the KeeSpace namespace class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addMember(self, name: str, value: Any, **kwargs) -> None:
    """
    Add a member to the KeeSpace namespace.
    """
    if value is None:
      if kwargs.get('_recursion2', False):
        raise RecursionError
      return self._addMember(name, name, _recursion2=True)
    valueType = self._getValueType()
    if valueType is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self._setValueType(type(value))
      return self._addMember(name, value, _recursion=True)
    else:
      if not isinstance(value, valueType):
        es = """Found inconsistent value type for members of the '%s' class 
        under construction. Member named: '%s' has value type: '%s', 
        while the previous values had value type: '%s'. Values passed to 
        'auto' must be of the same type for the same enumeration!"""
        clsName = self.space.getClassName()
        actName = type(value).__name__
        expName = valueType.__name__
        e = es % (clsName, name, actName, expName)
        raise NotImplementedError(e)
    existing = self.space.get('__future_entries__', {})
    if name in existing:
      raise NotImplementedError('duplicate keenum: %s' % name)
    existing[name] = value
    self.space['__future_entries__'] = existing

  def _setValueType(self, valueType: type, ) -> None:
    """
    Set the value type for the KeeSpace namespace.
    """
    if '__future_value_type__' in self.space:
      existingType = self.space['__future_value_type__']
      if valueType is not existingType:
        es = """Found inconsistent value type for members of the '%s' class 
        under construction. The value type was previously set to: '%s', 
        while the new value type is: '%s'. Values passed to 'auto' must be 
        of the same type for the same enumeration!"""
        clsName = self.space.getClassName()
        e = es % (clsName, existingType.__name__, valueType.__name__)
        raise NotImplementedError(e)
      return
    self.space['__future_value_type__'] = valueType

  def _getValueType(self) -> type:
    """
    Get the value type for the KeeSpace namespace.
    """
    return self.space.get('__future_value_type__', None)

  def setItemHook(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """
    The setItemHook method is called when an item is set in the
    namespace.
    """
    if isinstance(value, AbstractObject):
      if hasattr(value, '__auto_member__'):
        if getattr(value, '__auto_member__'):
          if TYPE_CHECKING:
            assert isinstance(value, _AutoMember)
          self._addMember(key.upper(), value.getValue())
          return True
    return False

  def preCompileHook(self, compiledSpace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    futureEntries = compiledSpace.get('__future_entries__', None)
    if futureEntries is None:
      compiledSpace['__future_entries__'] = dict()
    return compiledSpace



"""
_FieldProps holds the many properties used by the Field class. This
private class is intended solely to avoid excessive amounts of code in a
single file.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from functools import wraps

from ..parse import maybe
from ..static import AbstractObject
from ..text import typeMsg
from ..waitaminute import MissingVariable, TypeException, VariableNotNone

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

from types import FunctionType
from types import FunctionType as Func
from types import MethodType
from types import MethodType as Meth

if TYPE_CHECKING:
  from typing import Any, Self, Union, TypeAlias, Never, Callable


#
# def _flexSet(setterFunction: Callable) -> Callable:
#   """
#   Wraps setter functions calling them first with both new and old values,
#   then if raises 'TypeError', calls again with only the new value.
#   """
#
#   @wraps(setterFunction)
#   def wrapped(self: Any, value: Any, oldValue: Any = None) -> None:
#     """
#     Calls the setter function with both new and old values, then if
#     raises 'TypeError', calls again with only the new value.
#     """
#     try:
#       out = setterFunction(self, value, oldValue)
#     except TypeError as typeError:
#       if 'positional' in str(typeError):
#         try:
#           out = setterFunction(self, value)
#         except Exception as exception:
#           raise exception from typeError
#         else:
#           return out
#         finally:
#           pass
#       else:
#         raise typeError
#     else:
#       return out
#     finally:
#       pass
#
#   if isinstance(wrapped, Func):
#     return wrapped
#   raise TypeException('wrapped', wrapped, Func)
#
#
# def _flexDelete(deleterFunction: Callable) -> Callable:
#   """
#   Wraps deleter functions calling them first with the old value, then if
#   raises 'TypeError', calls again without the old value.
#   """
#
#   @wraps(deleterFunction)
#   def wrapped(self: Any, oldValue: Any = None) -> None:
#     """
#     Calls the deleter function with the old value, then if raises
#     'TypeError', calls again without the old value.
#     """
#     try:
#       out = deleterFunction(self, oldValue)
#     except TypeError as typeError:
#       if 'positional' in str(typeError):
#         try:
#           out = deleterFunction(self)
#         except Exception as exception:
#           raise exception from typeError
#         else:
#           return out
#         finally:
#           pass
#       else:
#         raise typeError
#     else:
#       return out
#     finally:
#       pass
#
#   if isinstance(wrapped, Func):
#     return wrapped
#   raise TypeException('wrapped', wrapped, Func)


class AbstractDescriptor(AbstractObject):
  """
  AbstractField provides an implementation of the descriptor protocol
  that allow the owning class to explicitly define the accessor methods.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables

  #  Private variables
  #  # Function keys allowing a subclass to override
  #  #  # Accessors
  __getter_key__ = None  # Get
  __setter_keys__ = None  # Set
  __deleter_keys__ = None  # Delete
  #  #  # Notifiers
  __pre_get_keys__ = None  # Before get
  __post_get_keys__ = None  # After get (using try/finally)
  __pre_set_keys__ = None  # Before set
  __post_set_keys__ = None  # After set
  __pre_delete_keys__ = None  # Before delete
  __post_delete_keys__ = None  # After delete
  #  # Function objects
  #  #  # Accessors
  __getter_func__ = None  # Get
  __setter_funcs__ = None  # Set
  __deleter_funcs__ = None  # Delete
  #  #  # Notifiers
  __pre_get_funcs__ = None  # Before get
  __post_get_funcs__ = None  # After get (using try/finally)
  __pre_set_funcs__ = None  # Before set
  __post_set_funcs__ = None  # After set
  __pre_delete_funcs__ = None  # Before delete
  __post_delete_funcs__ = None  # After delete

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Getter for the names of the accessors

  def _getGetKey(self, ) -> str:
    """
    Getter-function for the name of the single and required method called
    to retrieve the value to be returned by '__get__'. The recommended
    design pattern is to decorate a method in the class body with 'GET'.
    The decorated method should simply be an instance method.
    """
    if self.__getter_key__ is None:
      raise MissingVariable('__getter_key__', str)
    if isinstance(self.__getter_key__, str):
      return self.__getter_key__
    name, value = '__getter_key__', self.__getter_key__
    raise TypeException(name, value, str)

  def _getSetKeys(self, **kwargs) -> tuple[str, ...]:
    """
    Getter-function for the names of the methods called to set the value
    of the field. The recommended design pattern is to decorate methods
    in the class body with 'SET'. The decorated methods should simply be
    instance methods.
    """
    if self.__setter_keys__ is None:
      return ()
    if isinstance(self.__setter_keys__, list):
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.__setter_keys__ = (*self.__setter_keys__,)
      return self._getSetKeys(_recursion=True, )
    if isinstance(self.__setter_keys__, tuple):
      for key in self.__setter_keys__:
        if not isinstance(key, str):
          raise TypeError(typeMsg('setterKey', key, str))
      else:
        return self.__setter_keys__
    name, value = '__setter_keys__', self.__setter_keys__
    raise TypeException(name, value, tuple)

  def _getDeleteKeys(self, **kwargs) -> tuple[str, ...]:
    """
    Getter-function for the names of the methods called to delete the
    field. The recommended design pattern is to decorate methods in the
    class body with 'DELETE'. The decorated methods should simply be
    instance methods.
    """
    if self.__deleter_keys__ is None:
      return ()
    if isinstance(self.__deleter_keys__, list):
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.__deleter_keys__ = (*self.__deleter_keys__,)
      return self._getDeleteKeys(_recursion=True, )
    if isinstance(self.__deleter_keys__, tuple):
      for key in self.__deleter_keys__:
        if not isinstance(key, str):
          raise TypeError(typeMsg('deleterKey', key, str))
      else:
        return self.__deleter_keys__
    name, value = '__deleter_keys__', self.__deleter_keys__
    raise TypeException(name, value, tuple)

  def _getPreGetKeys(self) -> tuple[str, ...]:
    """
    Getter-function for the names of the methods called before '__get__'
    returns. """
    return maybe(self.__pre_get_keys__, ())

  def _getPostGetKeys(self) -> tuple[str, ...]:
    """
    Getter-function for keys to functions called after '__get__' has
    returned, using try/finally.
    """
    return maybe(self.__post_get_keys__, ())

  def _getPreSetKeys(self) -> tuple[str, ...]:
    """
    Getter-function for keys to functions called before '__set__' returns
    """
    return maybe(self.__pre_set_keys__, ())

  def _getPostSetKeys(self) -> tuple[str, ...]:
    """
    Getter-function for keys to functions called after '__set__' has
    returned, using try/finally.
    """
    return maybe(self.__post_set_keys__, ())

  def _getPreDeleteKeys(self) -> tuple[str, ...]:
    """
    Getter-function for keys to functions called before '__delete__'
    returns.
    """
    return maybe(self.__pre_delete_keys__, ())

  def _getPostDeleteKeys(self) -> tuple[str, ...]:
    """
    Getter-function for keys to functions called after '__delete__' has
    returned, using try/finally.
    """
    return maybe(self.__post_delete_keys__, ())

  #  Getter for accessor functions directly

  def _getGet(self, **kwargs) -> Func:
    """
    Getter-function for the getter function.

    This function receives the current instance as the single argument:
    func(self.instance) -> Any

    """
    if self.owner is self.__field_owner__:  # return function object
      if self.__getter_func__ is not None:
        return self._getGetFuncObject(**kwargs)
    getterKey = self._getGetKey()
    getterFunc = getattr(self.owner, getterKey)
    if getterFunc is None:
      raise MissingVariable(getterKey, Func)
    if isinstance(getterFunc, Func):
      return getterFunc
    name, value = getterKey, getterFunc
    raise TypeException(name, value, Func)

  def _getGetFuncObject(self, **kwargs) -> Func:
    """
    Getter-function for the getter function object.
    """
    if self.__getter_func__ is not None:
      if isinstance(self.__getter_func__, Func):
        return self.__getter_func__
      name, value = '__getter_func__', self.__getter_func__
      raise TypeException(name, value, Func)
    if kwargs.get('_recursion', False):
      raise RecursionError
    self.__getter_func__ = self._getGet()
    return self._getGetFuncObject(_recursion=True, )

  def _getSet(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the setter functions.

    These functions receive:
    func(self.instance, value: Any) -> None
    """
    if self.owner is self.__field_owner__:  # return function objects
      if self.__setter_funcs__ is not None:
        return self._getSetFuncObjects(**kwargs)
    setterKeys = self._getSetKeys(**kwargs)
    setterFuncs = []
    for setterKey in setterKeys:
      setterFunc = getattr(self.owner, setterKey, None)
      if setterFunc is None:
        raise MissingVariable(setterKey, Func)
      if isinstance(setterFunc, Func):
        setterFuncs.append(setterFunc)
        continue
      name, value = setterKey, setterFunc
      raise TypeException(name, value, Func)
    return (*setterFuncs,)

  def _getSetFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the setter function objects.
    """
    if self.__setter_funcs__ is not None:
      if isinstance(self.__setter_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__setter_funcs__ = (*self.__setter_funcs__,)
        return self._getSet(_recursion=True, )
      if isinstance(self.__setter_funcs__, tuple):
        for setterFunc in self.__setter_funcs__:
          if not isinstance(setterFunc, Func):
            name, value = '__setter_funcs__', setterFunc
            raise TypeException(name, value, Func)
        else:
          return self.__setter_funcs__
      if isinstance(self.__setter_funcs__, Func):
        return (self.__setter_funcs__,)
      name, value = '__setter_funcs__', self.__setter_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getDelete(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the deleter functions.

    These functions receive:
    func(self.instance) -> None
    """
    if self.owner is self.__field_owner__:  # return function objects
      if self.__deleter_funcs__ is not None:
        return self._getDeleteFuncObjects(**kwargs)
    deleterKeys = self._getDeleteKeys(**kwargs)
    deleterFuncs = []
    for deleterKey in deleterKeys:
      deleterFunc = getattr(self.owner, deleterKey, None)
      if deleterFunc is None:
        raise MissingVariable(deleterKey, Func)
      if isinstance(deleterFunc, Func):
        deleterFuncs.append(deleterFunc)
        continue
      name, value = deleterKey, deleterFunc
      raise TypeException(name, value, Func)
    return (*deleterFuncs,)

  def _getDeleteFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the deleter function objects.
    """
    if self.__deleter_funcs__ is not None:
      if isinstance(self.__deleter_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__deleter_funcs__ = (*self.__deleter_funcs__,)
        return self._getDelete(_recursion=True, )
      if isinstance(self.__deleter_funcs__, tuple):
        for deleterFunc in self.__deleter_funcs__:
          if not isinstance(deleterFunc, Func):
            name, value = '__deleter_funcs__', deleterFunc
            raise TypeException(name, value, Func)
        else:
          return self.__deleter_funcs__
      if isinstance(self.__deleter_funcs__, Func):
        return (self.__deleter_funcs__,)
      name, value = '__deleter_funcs__', self.__deleter_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getPreGet(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the pre-get functions.

    These functions are called before the __get__ returns, but after the
    return value has been determined including any side effects. These
    functions are then called with:
    func(self.instance, value: Any) -> None
    """
    if self.owner is self.__field_owner__:
      # return function objects
      if self.__pre_get_funcs__ is not None:
        return self._getPreGetFuncObjects(**kwargs)
    preGetKeys = self._getPreGetKeys()
    preGetFuncs = []
    for preGetKey in preGetKeys:
      preGetFunc = getattr(self.owner, preGetKey, None)
      if preGetFunc is None:
        raise MissingVariable(preGetKey, Func)
      if isinstance(preGetFunc, Func):
        preGetFuncs.append(preGetFunc)
        continue
      name, value = preGetKey, preGetFunc
      raise TypeException(name, value, Func)
    return (*preGetFuncs,)

  def _getPreGetFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the pre-get function objects.
    """
    if self.__pre_get_funcs__ is not None:
      if isinstance(self.__pre_get_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__pre_get_funcs__ = (*self.__pre_get_funcs__,)
        return self._getPreGet(_recursion=True, )
      if isinstance(self.__pre_get_funcs__, tuple):
        for preGetFunc in self.__pre_get_funcs__:
          if not isinstance(preGetFunc, Func):
            name, value = '__pre_get_funcs__', preGetFunc
            raise TypeException(name, value, Func)
        else:
          return self.__pre_get_funcs__
      if isinstance(self.__pre_get_funcs__, Func):
        return (self.__pre_get_funcs__,)
      name, value = '__pre_get_funcs__', self.__pre_get_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getPostGet(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the post-get functions.

    These functions are called after the __get__ has returned:
    try:
      return value
    finally:
      func(self.instance, value: Any) -> None
    """
    if self.owner is self.__field_owner__:  # return function objects
      if self.__post_get_funcs__ is not None:
        return self._getPostGetFuncObjects(**kwargs)
    postGetKeys = self._getPostGetKeys()
    postGetFuncs = []
    for postGetKey in postGetKeys:
      postGetFunc = getattr(self.owner, postGetKey, None)
      if postGetFunc is None:
        raise MissingVariable(postGetKey, Func)
      if isinstance(postGetFunc, Func):
        postGetFuncs.append(postGetFunc)
        continue
      name, value = postGetKey, postGetFunc
      raise TypeException(name, value, Func)
    return (*postGetFuncs,)

  def _getPostGetFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the post-get function objects.
    """
    if self.__post_get_funcs__ is not None:
      if isinstance(self.__post_get_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__post_get_funcs__ = (*self.__post_get_funcs__,)
        return self._getPostGet(_recursion=True, )
      if isinstance(self.__post_get_funcs__, tuple):
        for postGetFunc in self.__post_get_funcs__:
          if not isinstance(postGetFunc, Func):
            name, value = '__post_get_funcs__', postGetFunc
            raise TypeException(name, value, Func)
        else:
          return self.__post_get_funcs__
      if isinstance(self.__post_get_funcs__, Func):
        return (self.__post_get_funcs__,)
      name, value = '__post_get_funcs__', self.__post_get_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getPreSet(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the pre-set hooks.

    These functions are called before __set__ returns with:
    oldValue = __get__(...)  # The current value or exception
    func(self.instance, value: Any, oldValue: Any) -> None
    """
    if self.owner is self.__field_owner__:
      # return function objects
      if self.__pre_set_funcs__ is not None:
        return self._getPreSetFuncObjects(**kwargs)
    preSetKeys = self._getPreSetKeys()
    preSetFuncs = []
    for preSetKey in preSetKeys:
      preSetFunc = getattr(self.owner, preSetKey, None)
      if preSetFunc is None:
        raise MissingVariable(preSetKey, Func)
      if isinstance(preSetFunc, Func):
        preSetFuncs.append(preSetFunc)
        continue
      name, value = preSetKey, preSetFunc
      raise TypeException(name, value, Func)
    return (*preSetFuncs,)

  def _getPreSetFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the pre-set function objects.
    """
    if self.__pre_set_funcs__ is not None:
      if isinstance(self.__pre_set_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__pre_set_funcs__ = (*self.__pre_set_funcs__,)
        return self._getPreSet(_recursion=True, )
      if isinstance(self.__pre_set_funcs__, tuple):
        for preSetFunc in self.__pre_set_funcs__:
          if not isinstance(preSetFunc, Func):
            name, value = '__pre_set_funcs__', preSetFunc
            raise TypeException(name, value, Func)
        else:
          return self.__pre_set_funcs__
      if isinstance(self.__pre_set_funcs__, Func):
        return (self.__pre_set_funcs__,)
      name, value = '__pre_set_funcs__', self.__pre_set_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getPostSet(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the post-set hooks.

    These functions are called after __set__ has returned:
    oldValue = __get__(...)  # The current value or exception
    __set__(...)  # The new value is set
    newValue = __get__(...)  # The new value retrieved through __get__
    __set__(...)
    func(self.instance, newValue: Any, oldValue: Any) -> None
    """
    if self.owner is self.__field_owner__:  # return function objects
      if self.__post_set_funcs__ is not None:
        return self._getPostSetFuncObjects(**kwargs)
    postSetKeys = self._getPostSetKeys()
    postSetFuncs = []
    for postSetKey in postSetKeys:
      postSetFunc = getattr(self.owner, postSetKey, None)
      if postSetFunc is None:
        raise MissingVariable(postSetKey, Func)
      if isinstance(postSetFunc, Func):
        postSetFuncs.append(postSetFunc)
        continue
      name, value = postSetKey, postSetFunc
      raise TypeException(name, value, Func)
    return (*postSetFuncs,)

  def _getPostSetFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the post-set function objects.
    """
    if self.__post_set_funcs__ is not None:
      if isinstance(self.__post_set_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__post_set_funcs__ = (*self.__post_set_funcs__,)
        return self._getPostSet(_recursion=True, )
      if isinstance(self.__post_set_funcs__, tuple):
        for postSetFunc in self.__post_set_funcs__:
          if not isinstance(postSetFunc, Func):
            name, value = '__post_set_funcs__', postSetFunc
            raise TypeException(name, value, Func)
        else:
          return self.__post_set_funcs__
      if isinstance(self.__post_set_funcs__, Func):
        return (self.__post_set_funcs__,)
      name, value = '__post_set_funcs__', self.__post_set_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getPreDelete(self, **kwargs, ) -> tuple[Func, ...]:
    """
    Getter-function for the pre-delete hooks.

    These functions are called before __delete__ returns with:
    oldValue = __get__(...)  # The current value

    Please note that if oldValue is the 'DELETED' sentinel value or the
    __get__ raises AttributeError, then the error is expected to
    propagate. The same is not True for the set methods.

    func(self.instance, oldValue) -> None
    """
    if self.owner is self.__field_owner__:  # return function objects
      if self.__pre_delete_funcs__ is not None:
        return self._getPreDeleteFuncObjects(**kwargs)
    preDeleteKeys = self._getPreDeleteKeys()
    preDeleteFuncs = []
    for preDeleteKey in preDeleteKeys:
      preDeleteFunc = getattr(self.owner, preDeleteKey, None)
      if preDeleteFunc is None:
        raise MissingVariable(preDeleteKey, Func)
      if isinstance(preDeleteFunc, Func):
        preDeleteFuncs.append(preDeleteFunc)
        continue
      name, value = preDeleteKey, preDeleteFunc
      raise TypeException(name, value, Func)
    return (*preDeleteFuncs,)

  def _getPreDeleteFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the pre-delete function objects.
    """
    if self.__pre_delete_funcs__ is not None:
      if isinstance(self.__pre_delete_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__pre_delete_funcs__ = (*self.__pre_delete_funcs__,)
        return self._getPreDelete(_recursion=True, )
      if isinstance(self.__pre_delete_funcs__, tuple):
        for preDeleteFunc in self.__pre_delete_funcs__:
          if not isinstance(preDeleteFunc, Func):
            name, value = '__pre_delete_funcs__', preDeleteFunc
            raise TypeException(name, value, Func)
        else:
          return self.__pre_delete_funcs__
      if isinstance(self.__pre_delete_funcs__, Func):
        return (self.__pre_delete_funcs__,)
      name, value = '__pre_delete_funcs__', self.__pre_delete_funcs__
      raise TypeException(name, value, tuple)
    return ()

  def _getPostDelete(self, **kwargs, ) -> tuple[Func, ...]:
    """
    Getter-function for the post-delete hooks.

    These functions are called after __delete__ has returned:
    oldValue = __get__(...)  # The current value
    __delete__(...)  # The value is deleted
    func(self.instance, oldValue) -> None

    It is possible for a post delete hook to restore the value at this
    stage, although doing so will break the expectation that the following
    should raise AttributeError:

    foo.bar = 69
    del foo.bar
    foo.bar  # Raises AttributeError
    """
    if self.owner is self.__field_owner__:  # return function objects
      if self.__post_delete_funcs__ is not None:
        return self._getPostDeleteFuncObjects(**kwargs)
    postDeleteKeys = self._getPostDeleteKeys()
    postDeleteFuncs = []
    for postDeleteKey in postDeleteKeys:
      postDeleteFunc = getattr(self.owner, postDeleteKey, None)
      if postDeleteFunc is None:
        raise MissingVariable(postDeleteKey, Func)
      if isinstance(postDeleteFunc, Func):
        postDeleteFuncs.append(postDeleteFunc)
        continue
      name, value = postDeleteKey, postDeleteFunc
      raise TypeException(name, value, Func)
    return (*postDeleteFuncs,)

  def _getPostDeleteFuncObjects(self, **kwargs) -> tuple[Func, ...]:
    """
    Getter-function for the post-delete function objects.
    """
    if self.__post_delete_funcs__ is not None:
      if isinstance(self.__post_delete_funcs__, list):
        if kwargs.get('_recursion', False):
          raise RecursionError
        self.__post_delete_funcs__ = (*self.__post_delete_funcs__,)
        return self._getPostDelete(_recursion=True, )
      if isinstance(self.__post_delete_funcs__, tuple):
        for postDeleteFunc in self.__post_delete_funcs__:
          if not isinstance(postDeleteFunc, Func):
            name, value = '__post_delete_funcs__', postDeleteFunc
            raise TypeException(name, value, Func)
        else:
          return self.__post_delete_funcs__
      if isinstance(self.__post_delete_funcs__, Func):
        return (self.__post_delete_funcs__,)
      name, value = '__post_delete_funcs__', self.__post_delete_funcs__
      raise TypeException(name, value, tuple)
    return ()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _setGetter(self, callMeMaybe: Func) -> Func:
    """
    Setter-function for the getter function.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_setGetter', callMeMaybe
      raise TypeException(name, value, Func)
    self.__getter_func__ = callMeMaybe
    self.__getter_key__ = callMeMaybe.__name__
    return callMeMaybe

  def _addSetter(self, callMeMaybe: Func) -> Func:
    """
    Add a setter function to the field.

    This function is used to add a setter function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addSetter', callMeMaybe
      raise TypeException(name, value, Func)
    existingFuncs = self._getSet()
    existingKeys = self._getSetKeys()
    self.__setter_funcs__ = (*existingFuncs, callMeMaybe)
    self.__setter_keys__ = (*existingKeys, callMeMaybe.__name__,)
    return callMeMaybe

  def _addDeleter(self, callMeMaybe: Func) -> Func:
    """
    Add a deleter function to the field.

    This function is used to add a deleter function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addDeleter', callMeMaybe
      raise TypeException(name, value, Func)
    existingFuncs = self._getDelete()
    existingKeys = self._getDeleteKeys()
    self.__deleter_funcs__ = (*existingFuncs, callMeMaybe)
    self.__deleter_keys__ = (*existingKeys, callMeMaybe.__name__,)
    return callMeMaybe

  def _addPreGet(self, callMeMaybe: Func) -> Func:
    """
    Add a pre-get function to the field.

    This function is used to add a pre-get function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addPreGet', callMeMaybe
      raise TypeException(name, value, Func)
    existingKeys = self._getPreGetKeys()
    existingFuncs = self._getPreGetFuncs()
    self.__pre_get_keys__ = (*existingKeys, callMeMaybe.__name__,)
    self.__pre_get_funcs__ = (*existingFuncs, callMeMaybe)
    return callMeMaybe

  def _addPostGet(self, callMeMaybe: Func) -> Func:
    """
    Add a post-get function to the field.

    This function is used to add a post-get function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addPostGet', callMeMaybe
      raise TypeException(name, value, Func)
    existingKeys = self._getPostGetKeys()
    existingFuncs = self._getPostGet()
    self.__post_get_keys__ = (*existingKeys, callMeMaybe.__name__,)
    self.__post_get_funcs__ = (*existingFuncs, callMeMaybe)
    return callMeMaybe

  def _addPreSet(self, callMeMaybe: Func) -> Func:
    """
    Add a pre-set function to the field.

    This function is used to add a pre-set function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addPreSet', callMeMaybe
      raise TypeException(name, value, Func)
    existingFuncs = self._getPreSet()
    existingKeys = self._getPreSetKeys()
    self.__pre_set_keys__ = (*existingKeys, callMeMaybe.__name__,)
    self.__pre_set_funcs__ = (*existingFuncs, callMeMaybe)
    return callMeMaybe

  def _addPostSet(self, callMeMaybe: Func) -> Func:
    """
    Add a post-set function to the field.

    This function is used to add a post-set function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addPostSet', callMeMaybe
      raise TypeException(name, value, Func)
    existingFuncs = self._getPostSet()
    existingKeys = self._getPostSetKeys()
    self.__post_set_keys__ = (*existingKeys, callMeMaybe.__name__,)
    self.__post_set_funcs__ = (*existingFuncs, callMeMaybe)
    return callMeMaybe

  def _addPreDelete(self, callMeMaybe: Func) -> Func:
    """
    Add a pre-delete function to the field.

    This function is used to add a pre-delete function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addPreDelete', callMeMaybe
      raise TypeException(name, value, Func)
    existingKeys = self._getPreDeleteKeys()
    existingFuncs = self._getPreDelete()
    self.__pre_delete_keys = (*existingKeys, callMeMaybe.__name__,)
    self.__pre_delete_funcs__ = (*existingFuncs, callMeMaybe)
    return callMeMaybe

  def _addPostDelete(self, callMeMaybe: Func) -> Func:
    """
    Add a post-delete function to the field.

    This function is used to add a post-delete function to the field. The
    function should be an instance method of the owner class.
    """
    if not isinstance(callMeMaybe, Func):
      name, value = '_addPostDelete', callMeMaybe
      raise TypeException(name, value, Func)
    existingKeys = self._getPostDeleteKeys()
    existingFuncs = self._getPostDelete()
    self.__post_delete_keys__ = (*existingKeys, callMeMaybe.__name__,)
    self.__post_delete_funcs__ = (*existingFuncs, callMeMaybe)
    return callMeMaybe



"""The 'worktoy.attr' module implements the descriptor protocol."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._field_properties import _FieldProperties
from ._abstract_descriptor import AbstractDescriptor
from . import accessor_hooks
from ._field import Field
from ._wait_for_it import WaitForIt
from ._abstract_box import AbstractBox
from ._attri_box import AttriBox
from ._flex_box import FlexBox

__all__ = [
    'accessor_hooks',
    'AbstractDescriptor',
    'Field',
    'WaitForIt',
    'AbstractBox',
    'AttriBox',
    'FlexBox',
]



"""THIS is the token object indicating a class before it is created. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from worktoy.waitaminute import IllegalInstantiationError

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Never


class Zeroton(type):
  """Metaclass ensuring the 'THIS' class never instantiates."""

  def __new__(mcls, *args, **kwargs) -> type:
    """Prevent instantiation of THIS class."""
    bases = (object,)
    name = args[0]
    key = '__%s__' % name
    space = dict()
    space[key] = True
    return type.__new__(mcls, name, bases, space)

  def __call__(cls, *args, **kwargs) -> Never:
    """Prevent instantiation of THIS class."""
    raise IllegalInstantiationError(cls)

  def __str__(cls, ) -> str:
    """Return the string representation of THIS class."""
    info = """The '%s' class provides a placeholder for classes not yet 
    created. """
    return info % cls.__name__

  def __hash__(cls) -> int:
    """Return the hash of THIS class."""
    return 69420


class THIS(metaclass=Zeroton):
  """THIS is the token object indicating a class before it is created. """
  pass


class OWNER(metaclass=Zeroton):
  """OWNER is the token object indicating the owner of the descriptor. """
  pass


class ATTR(metaclass=Zeroton):
  """ATTR is the token object indicating the descriptor itself. """
  pass



"""
The 'worktoy.attr.accessor_hooks' module provides notifying hooks for the
descriptors based on the 'worktoy.attr.AbstractDescriptor' base class and
its subclasses.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._priority import _Priority  # Private
from ._abstract_descriptor_hook import AbstractDescriptorHook

__all__ = [
    'AbstractDescriptorHook',
]



"""_CurrentOwner is a private method used by AbstractObject to expose
the current owner of the descriptor instance."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..waitaminute import MissingVariable

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Never
  from . import AbstractObject


class _CurrentOwner:
  """_CurrentOwner is a private method used by AbstractObject to expose
  the current owner of the descriptor instance."""

  def __get__(self, instance: Any, owner: type) -> Any:
    """Return the current owner of the descriptor instance."""
    if instance is None:
      return self
    if TYPE_CHECKING:
      assert isinstance(instance, AbstractObject)
    if instance.__current_owner__ is None:
      if instance.__field_owner__ is None:
        return None
      return instance.__field_owner__
    return instance.__current_owner__

  def __set__(self, *_) -> Never:
    """This should never happen."""
    raise RuntimeError

  def __delete__(self, *_) -> Never:
    """This should never happen."""
    raise RuntimeError



"""BaseMeta provides a basic metaclass. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import AbstractMetaclass
from . import BaseSpace as BSpace
from . import Types

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  pass


class BaseMeta(AbstractMetaclass):
  """BaseMeta provides a basic metaclass."""

  @classmethod
  def __prepare__(mcls, name: str, bases: Types, **kwargs) -> BSpace:
    """Prepare the class namespace."""
    return BSpace(mcls, name, bases, **kwargs)



"""BaseObject exposes the functionality of the custom metaclass
implementations in the 'worktoy' library. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import BaseMeta
from ..static import AbstractObject

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


class BaseObject(AbstractObject, metaclass=BaseMeta):
  """BaseObject exposes the functionality of the custom metaclass
  implementations in the 'worktoy' library. """



"""_CurrentInstance is a private method used by AbstractObject to expose
the current owning instance of the descriptor instance."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Never

  from . import AbstractObject


class _CurrentInstance:
  """_CurrentInstance is a private method used by AbstractObject to
  expose
  the current owning instance of the descriptor instance."""

  def __get__(self, desc: Any, owner: type) -> Any:
    """Return the current owning instance of the descriptor instance."""
    if desc is None:
      return self
    if TYPE_CHECKING:
      assert isinstance(desc, AbstractObject)
    return desc.__current_instance__

  def __set__(self, *_) -> Never:
    """This should never happen."""
    raise RuntimeError

  def __delete__(self, *_) -> Never:
    """This should never happen."""
    raise RuntimeError



"""
InstanceException is a custom exception raised when a descriptor may be
accessed only through the owning class, but where __get__ receives as
instance argument something other than None.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..text import joinWords, monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any

  from ..attr import _FieldProperties as Desc


class InstanceException(ValueError):
  """
  InstanceException is a custom exception raised when a descriptor may be
  accessed only through the owning class, but where __get__ receives as
  instance argument something other than None.
  """

  descObject = _Attribute()
  descType = _Attribute()
  instance = _Attribute()
  owner = _Attribute()
  fieldName = _Attribute()

  def __init__(self, desc: Desc, ins: Any, ) -> None:
    """Initialize the InstanceException object."""
    self.descObject = desc
    self.descType = type(desc)
    self.instance = ins
    self.owner = type(ins)
    self.fieldName = desc.__field_name__
    infoSpec = """
    '%s' objects may only be accessed through the owning class, 
    not through instances!
    """
    info = monoSpace(infoSpec % (type(desc).__name__,))
    ValueError.__init__(self, info)

  def __eq__(self, other: Any) -> bool:
    """Compare the InstanceException object with another object."""
    cls = type(self)
    if not isinstance(other, cls):
      return False
    if other is self:
      return True
    if self.owner is not other.owner:
      return False
    if self.fieldName != other.fieldName:
      return False
    return True



"""
Waila provides a well formatted summary of positional and keyword
arguments. The following attributes specify the formatting:

- header: str
The header above the summary.

- lineLength=77: int
The length of each line in the summary.

- indent=2: int
The number of spaces to indent each line.

- lineSep=os.linesep: str
The separator between lines in the summary. Note the default value is left
up to the os.linesep.

- columnSep='|': str
The separator between columns in the summary.


"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.parse import maybe

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self, TypeAlias

  Words: TypeAlias = list[str]
  Mappings: TypeAlias = dict[str, Any]


class _Value:
  """
  Primitive, private descriptor class
  """

  __fallback_key__ = None
  __value_key__ = None

  def __init__(self, valueKey: str, fallbackKey: str = None, ) -> None:
    self.__value_key__ = valueKey
    self.__fallback_key__ = fallbackKey

  def __get__(self, instance: Any, owner: type) -> Any:
    if instance is None:
      return self
    value = getattr(instance, self.__value_key__, None)
    fallback = getattr(instance, self.__fallback_key__, None)
    return maybe(value, fallback)


class Waila:
  """
  Waila provides a well formatted summary of positional and keyword
  arguments. The following attributes specify the formatting:

  - header: str
    The header above the summary.

  - lineLength=77: int
    The length of each line in the summary.

  - indent=2: int
    The number of spaces to indent each line.

  - lineSep=os.linesep: str
    The separator between lines in the summary. Note the default value is
    left
    up to the os.linesep.

  - columnSep='|': str
    The separator between columns in the summary.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback variables
  __fallback_length__ = 77
  __fallback_indent__ = '  '
  __fallback_line_sep__ = os.linesep
  __fallback_col_sep__ = ' | '
  __fallback_title__ = 'Arguments Summary'
  __fallback_entry__ = 'Index/Key'
  __fallback_type__ = 'type (__name__)'
  __fallback_arg__ = 'value (str)'

  #  Private variables
  __line_length__ = None
  __indent_spaces__ = None
  __line_sep__ = None
  __col_sep__ = None
  __title_text__ = None
  __pos_entry_header__ = None
  __pos_type_header__ = None
  __pos_arg_header__ = None
  __key_entry_header__ = None
  __key_type_header__ = None
  __key_arg_header__ = None
  __pos_args__ = None
  __key_args__ = None

  #  Public variables
  lineLength = _Value('__line_length__', '__fallback_length__', )
  indent = _Value('__indent_spaces__', '__fallback_indent__', )
  lineSep = _Value('__line_sep__', '__fallback_line_sep__', )
  columnSep = _Value('__col_sep__', '__fallback_col_sep__', )
  title = _Value('__title_text__', '__fallback_title__', )
  posEntryHeader = _Value('__pos_entry_header__', '__fallback_entry__', )
  posTypeHeader = _Value('__pos_type_header__', '__fallback_type__', )
  posArgHeader = _Value('__pos_arg_header__', '__fallback_arg__', )
  keyEntryHeader = _Value('__key_entry_header__', '__fallback_entry__', )
  keyTypeHeader = _Value('__key_type_header__', '__fallback_type__', )
  keyArgHeader = _Value('__key_arg_header__', '__fallback_arg__', )

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getPositionalArgs(self) -> Words:
    """Get the positional arguments as a list of strings."""
    return maybe(self.__pos_args__, [])

  def _getKeywordArgs(self) -> Mappings:
    """Get the keyword arguments as a dictionary of strings."""
    return maybe(self.__key_args__, {})

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _addPositionalArgs(self, *args: str) -> None:
    """
    Adds positional arguments to the Waila instance.
    """
    existing = self._getPositionalArgs()
    self.__pos_args__ = [*existing, *args, ]

  def _addKeywordArgs(self, **kwargs: Any) -> None:
    """
    Adds keyword arguments to the Waila instance.
    """
    existing = self._getKeywordArgs()
    for key, value in kwargs.items():
      if key not in existing:
        existing[key] = value
        continue
      raise NotImplementedError('duplicate key: %s' % key)
    self.__key_args__ = existing

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getEntryHeaderWidth(self, ) -> int:
    """
    Returns the width required by the entry header.
    """
    return max(len(self.posEntryHeader), len(self.keyEntryHeader))

  def _getPosEntrySpec(self) -> str:
    """
    Get the format specification for the positional entry in the summary.
    """
    args = self._getPositionalArgs() or ['']
    spec = """Index: %%%dd""" % len(str(len(args)))
    sample = spec % len(args)
    if len(sample) < self._getEntryHeaderWidth():
      return spec.rjust(self._getEntryHeaderWidth())
    return spec

  def _getKeyEntrySpec(self) -> str:
    """
    Get the format specification for the key entry in the summary.
    """
    args, kwargs = self._getPositionalArgs(), self._getKeywordArgs()
    keys = sorted([k for k, v in kwargs.items()], key=len) or ['']
    spec = """Key: %%%ds""" % len(keys[-1])
    sample = spec % keys[-1]
    if len(sample) < self._getEntryHeaderWidth():
      return spec.rjust(self._getEntryHeaderWidth())
    return spec

  def _getTypeHeaderWidth(self) -> int:
    """
    Returns the width required by the type header.
    """
    return max(len(self.posTypeHeader), len(self.keyTypeHeader))

  def _getTypeSpec(self, ) -> str:
    """
    Get the format specification for the type of the value.
    """
    args, kwargs = self._getPositionalArgs(), self._getKeywordArgs()
    argTypes = [type(arg).__name__ for arg in args]
    keyTypes = [type(v).__name__ for k, v in kwargs.items()]
    types = sorted([*argTypes, *keyTypes, ], key=len) or ['']

  def __call__(self, *args, **kwargs) -> str:
    """
    Call the Waila instance to get a formatted summary of positional and
    keyword arguments.
    """
    posEntrySpec = """Index: %%%dd""" % len(str(len(args)))
    keys = [k for k, v in kwargs.items()]
    longestKey = max([len(str(k)) for k in keys], default=0)
    keyEntryLength = max(len(posEntrySpec % len(args)), longestKey)
    keyEntrySpec = """Key: %%%ds""" % keyEntryLength



"""
CompoundException is a custom exception raised from multiple exceptions.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute
from ..parse import maybe
from ..text import joinWords, monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Self, Any, Iterator


class CompoundException(Exception):
  """
  CompoundException is a custom exception raised from multiple exceptions.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __exception_list__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  def _getExceptionList(self, ) -> tuple[Exception, ...]:
    """Getter-function for the exceptions."""
    return maybe(self.__exception_list__, ())

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *exceptions: Exception) -> None:
    """Initialize the CompoundException with the given exceptions."""
    self.__exception_list__ = exceptions
    infoSpec = """Encountered multiple exceptions: \n  %s"""
    excStr = '<br><tab>'.join(str(e) for e in self)
    info = monoSpace(infoSpec % excStr)
    Exception.__init__(self, info)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __iter__(self) -> Iterator[Exception]:
    """Iterate over the exceptions."""
    for exception in self._getExceptionList():
      yield exception



"""
'Attr' provides a descriptor exposing a private attribute to public access.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..static import AbstractObject
from ..waitaminute import MissingVariable, TypeException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self


class Attr(AbstractObject):
  """Creates a deferred function that is called when the __get__ is first
  called. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __private_name__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getPrivateName(self, **kwargs) -> str:
    """
    Getter-function for the private name.
    """
    if self.__private_name__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      name = AbstractObject.getPrivateName(self, )
      self._setPrivateName(name, )
      return self._getPrivateName(_recursion=True, )
    if isinstance(self.__private_name__, str):
      return self.__private_name__
    name, value = '__private_name__', self.__private_name__
    raise TypeException(name, value, str)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _setPrivateName(self, privateName: str) -> Self:
    """
    Setter-function for the private name.
    """
    if self.__private_name__ is not None:
      raise MissingVariable('__private_name__', str)
    if not isinstance(privateName, str):
      name, value = '_setPrivateName', privateName
      raise TypeException(name, value, str)
    self.__private_name__ = privateName
    return self

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, privateName: str = None) -> None:
    """
    Constructor for the Attr descriptor.
    """
    if privateName is not None:
      self._setPrivateName(privateName)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, **kwargs) -> Any:
    """
    Instance getter for the Attr descriptor.
    """
    pvtName = self._getPrivateName()
    try:
      out = getattr(self.instance, pvtName, )
    except Exception as exception:
      raise exception
    else:
      return out
    finally:
      pass

  def __instance_set__(self, value: Any, **kwargs) -> None:
    """
    Instance setter for the Attr descriptor.
    """
    pvtName = self._getPrivateName()
    try:
      oldValue = getattr(self.instance, pvtName, )
    except Exception as exception:
      if isinstance(exception, AttributeError):
        oldValue = None
      else:
        raise exception
    else:
      if oldValue is value:
        return
      setattr(self.instance, pvtName, value)
    finally:
      pass

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #



"""IllegalName is a custom exception raised to indicate that a namespace
entry is of a disallowed name. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..text import joinWords
from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Never


class IllegalName(Exception):
  """IllegalName is a custom exception raised to indicate that a namespace
  entry is of a disallowed name. """

  name = _Attribute()
  disallowedNames = _Attribute()

  def __init__(self, name: str, *disallowedNames) -> None:
    self.name = name
    self.disallowedNames = [*disallowedNames, ]
    if self.disallowedNames:
      infoSpec = """Tried setting name: '%s' which is on the list of 
      disallowed names: %s"""
      disallowedStr = joinWords(*self.disallowedNames, )
      info = infoSpec % (name, disallowedStr)
    else:
      infoSpec = """Tried setting disallowed name: '%s'!"""
      info = infoSpec % name
    Exception.__init__(self, info)



"""The Dispatch class dispatches a function call to the appropriate
function based on the type of the first argument. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import os

from . import AbstractObject
from ..static import TypeSig
from ..text import typeMsg, monoSpace
from ..waitaminute import HashMismatch, CastMismatch, ResolveException, \
  VariableNotNone, TypeException, CascadeException
from ..waitaminute import DispatchException

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Callable, TypeAlias, Never

  Types: TypeAlias = tuple[type, ...]
  Hashes: TypeAlias = list[int]
  HashMap: TypeAlias = dict[int, Callable]
  TypesMap: TypeAlias = dict[Types, Callable]
  CastMap: TypeAlias = dict[Types, Callable]
  CallMap: TypeAlias = dict[TypeSig, Callable]


class Dispatch(AbstractObject):
  """The Dispatch class dispatches a function call to the appropriate
  function based on the type of the first argument. """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __call_map__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getTypeSigs(self) -> list[TypeSig]:
    """Get the type signatures."""
    return [*self.__call_map__.keys(), ]

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, callMap: CallMap) -> None:
    self.__call_map__ = callMap

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """Set the name of the function."""
    AbstractObject.__set_name__(self, owner, name, **kwargs)

    for sig, call in self.__call_map__.items():
      if not isinstance(sig, TypeSig):
        raise TypeError(typeMsg('sig', sig, TypeSig))
      if not callable(call):
        from worktoy.mcls import FunctionType
        raise TypeError(typeMsg('call', call, FunctionType))
      TypeSig.replaceTHIS(sig, owner)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _fastCall(self, *args: Any, **kwargs: Any) -> Any:
    """Fast call the function."""
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.fast(*args)
      except HashMismatch:
        continue
      except TypeError as typeError:
        if 'required positional argument' in str(typeError):
          continue
        raise typeError
      except IndexError:
        continue
      except KeyError:
        continue
      if self.instance is not None:
        return call(self.instance, *posArgs, **kwargs)
      return call(*posArgs, **kwargs)
    raise DispatchException(self, *args)

  def _flexCall(self, *args: Any, **kwargs: Any) -> Any:
    """Flex call the function."""
    exceptions = []
    for sig, call in self.__call_map__.items():
      try:
        posArgs = sig.flex(*args)
      except Exception as exception:
        exceptions.append(exception)
        continue
      else:
        return self._fastCall(*posArgs, **kwargs)
    try:
      raise CascadeException(*exceptions)
    except CascadeException as cascadeException:
      raise DispatchException(self, *args) from cascadeException

  def _resolveArgs(self, *args) -> tuple:
    """Resolves tuples, lists and strings. """
    posArgs = []
    anyResolved = False
    for arg in args:
      if isinstance(arg, tuple):
        posArgs = [*posArgs, *arg]
        anyResolved = True
      elif isinstance(arg, list):
        posArgs = [*posArgs, *arg]
        anyResolved = True
      elif isinstance(arg, tuple):
        posArgs = [*posArgs, *arg]
        anyResolved = True
      elif isinstance(arg, str):
        evalArg = None
        try:
          evalArg = eval(arg)
        except NameError as nameError:
          evalArg = nameError
          posArgs = [*posArgs, arg]
        else:
          posArgs = [*posArgs, evalArg]
          anyResolved = True
        finally:
          pass
      else:
        posArgs.append(arg)
    if anyResolved:
      return (*posArgs,)
    raise ResolveException(self, args)

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    """Call the function."""
    try:
      return self._fastCall(*args, **kwargs)
    except DispatchException as fastException:
      pass
    try:
      return self._flexCall(*args, **kwargs)
    except DispatchException as flexException:
      pass
    try:
      posArgs = self._resolveArgs(*args)
    except ResolveException as resolveException:
      raise DispatchException(self, *args) from resolveException
    while True:
      try:
        posArgs = self._resolveArgs(*posArgs)
      except ResolveException:
        break
    return self.__call__(*posArgs, **kwargs)

  def __str__(self, ) -> str:
    """Get the string representation of the function."""
    sigStr = [str(sig) for sig in self.getTypeSigs()]
    info = """%s object supporting type signatures: \n%s"""
    sigLines = '<br><tab>'.join(sigStr)
    return monoSpace(info % (self.__field_name__, sigLines))

  def __repr__(self, ) -> str:
    """Get the string representation of the function."""
    return object.__repr__(self)



"""
AbstractHook provides an abstract baseclass for hooks used by the
namespaces in the metaclass system. These hooks allow modification of the
class creation process. The following static methods are expected from the
hooks:
- setItem: called before calls to __setitem__ on the namespace object.
- getItem: called before calls to __getitem__ on the namespace object.
- preCompile: called before the final namespace object is populated with
the conventional key, value pairs.
- postCompile: called after the final namespace object is populated with
the conventional key, value pairs.

Subclasses may implement either of the above methods. The default
implementations have no effect, so subclasses need only implement the
methods they are interested in.

AbstractHook implements the descriptor protocol such that calls to
'__get__' receive:
- instance: The current instance of the namespace object.
- owner: The current namespace class

Subclasses should be made available as attributes on the namespace subclass.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from ...static import AbstractObject, Alias
from ...waitaminute import MissingVariable, ReadOnlyError

try:
  from typing import TYPE_CHECKING, Type
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, Callable, Self, Never
  from worktoy.mcls import AbstractNamespace as ASpace

  AccessorHook = Callable[[ASpace, str, Any], Any]
  CompileHook = Callable[[ASpace, dict], dict]


class AbstractHook(AbstractObject):
  """
  AbstractHook provides an abstract baseclass for hooks used by the
  namespaces in the metaclass system. These hooks allow modification of the
  class creation process. The following static methods are expected from the
  hooks:
  - setItem: called before calls to __setitem__ on the namespace object.
  - getItem: called before calls to __getitem__ on the namespace object.
  - preCompile: called before the final namespace object is populated with
    the conventional key, value pairs.
  - postCompile: called after the final namespace object is populated with
    the conventional key, value pairs.

  Subclasses may implement either of the above methods. The default
  implementations have no effect, so subclasses need only implement the
  methods they are interested in.

  AbstractHook implements the descriptor protocol such that calls to
  '__get__' receive:
  - instance: The current instance of the namespace object.
  - owner: The current namespace class

  Subclasses should be made available as attributes on the namespace
  subclass.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Public variables
  space = Alias('instance')
  spaceClass = Alias('owner')

  #  TYPE_CHECKING
  if TYPE_CHECKING:
    from . import AbstractHook
    from .. import AbstractNamespace
    assert isinstance(AbstractHook.space, AbstractNamespace)
    assert issubclass(AbstractHook.spaceClass, AbstractNamespace)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  def getItemHook(self, key: str, value: Any, ) -> bool:
    """Hook for getItem. This is called before the __getitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """

  def setItemHook(self, key: str, value: Any, oldValue: Any, ) -> bool:
    """Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False. """

  def preCompileHook(self, compiledSpace: dict) -> dict:
    """Hook for preCompile. This is called before the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    return compiledSpace

  def postCompileHook(self, compiledSpace: dict) -> dict:
    """Hook for postCompile. This is called after the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged. """
    return compiledSpace

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """
    After the super call, adds one self to the namespace class as a hook
    class.
    """
    super().__set_name__(owner, name, **kwargs)
    self.spaceClass.addHook(self)



"""EZData leverages the 'worktoy' library to provide a dataclass."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ..mcls import FunctionType
from . import EZMeta

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False


def _root(callMeMaybe: FunctionType) -> FunctionType:
  """Root decorator for the EZData class."""
  setattr(callMeMaybe, '_root', True)
  return callMeMaybe


class EZData(metaclass=EZMeta):
  """EZData is a dataclass that provides a simple way to define data
  structures with validation and serialization capabilities. """

  if TYPE_CHECKING:
    def __iter__(self): pass

    def __next__(self): pass

    def __init__(self, *args, **kwargs): pass

    def __getitem__(self, key): pass

    def __setitem__(self, key, value): pass



"""
AbstractDescriptorHook defines the hooks surrounding the accessor
functions in the 'worktoy.attr.AbstractDescriptor' class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from warnings import warn

from ...static import AbstractObject
from ...text import stringList
from ...waitaminute import SubclassException

from .. import AbstractDescriptor

from . import _Priority

#  Below provides compatibility back to Python 3.7
try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, TypeAlias, Union


class AbstractDescriptorHook(AbstractObject):
  """
  AbstractDescriptorHook defines the hooks surrounding the accessor
  functions in the 'worktoy.attr.AbstractDescriptor' class.

  Note the 'None-means-no-change' convention: if a hook returns a value
  other than 'None', it replaces the value being handled by the accessor.
  Returning 'None' leaves the value unchanged. AbstractDescriptor explicitly
  disallows 'None' as a valid value, except for the above.

  The AbstractDescriptor class wraps three core accessor methods meant to
  be reimplemented by subclasses:

    Getter:  __instance_get__(self) -> Any
    Setter:  __instance_set__(self, value: Any) -> None
    Deleter: __instance_delete__(self) -> None

  Because hooks have access to the descriptor object, and the descriptor
  tracks both instance and owner, hook methods do not receive these as
  explicit arguments.

  Accessor control flows are defined as follows:

  Getter Flow:
    preValue = None
    for hook in hooks:
      result = __pre_get__(hook)
      if result is not None:
        if preValue is not None:
          raise HookConflict("Multiple pre-get overrides")
        preValue = result

    if preValue is None:
      value = __instance_get__(desc)
    else:
      value = preValue

    postValue = None
    for hook in hooks:
      result = __post_get__(hook, value)
      if result is not None:
        if postValue is not None:
          raise HookConflict("Multiple post-get overrides")
        postValue = result

    return postValue if postValue is not None else value

  Setter Flow:
    preValue = None
    for hook in hooks:
      result = __pre_set__(hook, value)
      if result is not None:
        if preValue is not None:
          raise HookConflict("Multiple pre-set overrides")
        preValue = result

    if preValue is not None:
      __instance_set__(desc, preValue)
      return

    try:
      oldVal = __instance_get__(desc)
    except MissingValue as missingValue:
      oldVal = missingValue

    postValue = None
    for hook in hooks:
      result = __post_set__(hook, value, oldVal)
      if result is not None:
        if postValue is not None:
          raise HookConflict("Multiple post-set overrides")
        postValue = result

    if postValue is not None:
      __instance_set__(desc, postValue)
    else:
      __instance_set__(desc, value)

  Deleter Flow:
    for hook in hooks:
      __pre_delete__(hook)

    try:
      oldVal = __instance_get__(desc)
    except MissingValue as missingValue:
      # Deletion of a missing value is forbidden; escalates to AttributeError
      raise AttributeError from missingValue  # Actual implementation may
      vary

    for hook in hooks:
      __post_delete__(hook, oldVal)

    __instance_delete__(desc)

  Where hooks may override values in the control flow at most one hook may
  provide a replacement value at each stage of the control flow. Multiple
  hooks providing replacement values will raise a HookConflict.

  Hooks apply according to their priority. The lower this priority value,
  the earlier the hook is applied in the control flow. The tie-breaker is
  deliberately random to avoid applying undefined deterministic order.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Fallback Variables

  #  Private Variables
  __priority_value__ = None

  #  Public Variables
  priority = _Priority()

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *args, **kwargs) -> None:
    keys = stringList("""priority, tier, rank""")
    kwargPriority, kwargs = self.parseKwargs(int, *keys, **kwargs)
    if kwargPriority is None:
      for arg in args:
        if isinstance(arg, int):
          kwargPriority = arg
          break
    else:
      self.priority = kwargPriority

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __pre_get__(self, ) -> Any:
    """
    Pre-get hook. Override this method to implement pre-get behavior.
    """

  def __post_get__(self, value: Any) -> Any:
    """
    Post-get hook. Override this method to implement post-get behavior.
    """

  def __pre_set__(self, value: Any, ) -> Any:
    """
    Pre-set hook. Override this method to implement pre-set behavior.
    """

  def __post_set__(self, value: Any, oldValue: Any) -> Any:
    """
    Post-set hook. Override this method to implement post-set behavior.
    """

  def __pre_delete__(self, ) -> None:
    """
    Pre-delete hook. Override this method to implement pre-delete behavior.
    """

  def __post_delete__(self, oldValue: Any) -> None:
    """
    Post-delete hook. Override this method to implement post-delete behavior.
    """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __set_name__(self, owner: type, name: str, **kwargs) -> None:
    """
    Sets the name of the hook in the owner class.
    """
    AbstractObject.__set_name__(self, owner, name, **kwargs)
    if issubclass(owner, AbstractDescriptor):
      owner.registerAccessorHook(name, self)
    else:
      raise SubclassException(owner, AbstractDescriptor, )

  def __int__(self, ) -> int:
    """
    Returns the priority value of the hook.
    """
    return self.priority.value

  def __str__(self) -> str:
    """
    Returns the string representation of the hook, which is its priority
    value.
    """
    warn('TODO: Implement __str__ for AbstractDescriptorHook', )
    return AbstractObject.__str__(self)

  def __repr__(self) -> str:
    """
    Returns the string representation of the hook, which is its priority
    value.
    """
    warn('TODO: Implement __repr__ for AbstractDescriptorHook', )
    return AbstractObject.__repr__(self)



"""
ArgumentException provides a custom exception raised when a function is
unable to parse received arguments.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from types import FunctionType as Func

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Optional, Self, TypeAlias


class ArgumentException(ValueError):
  """
  ArgumentException provides a custom exception raised when a function is
  unable to parse received arguments.
  """

  posArgs = _Attribute()
  keyArgs = _Attribute()
  func = _Attribute()

  def __init__(self, func: Func, *args: Any, **kwargs) -> None:
    self.posArgs = [*args, ]
    self.keyArgs = {**kwargs, }
    self.func = func
    ValueError.__init__(self, 'unable to parse arguments!')



"""CastException is a custom exception raised to indicate a failure for
the 'Cast' class to cast a value to a target type."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from . import _Attribute

try:
  from typing import TYPE_CHECKING
except ImportError:
  try:
    from typing_extensions import TYPE_CHECKING
  except ImportError:
    TYPE_CHECKING = False

if TYPE_CHECKING:
  from typing import Any, Self, Iterator

  from ..static import Cast


class CastException(TypeError):
  """CastException is a custom exception raised to indicate a failure for
  the 'Cast' class to cast a value to a target type."""

  value = _Attribute()
  target = _Attribute()
  why = _Attribute()
  cast = _Attribute()

  def __init__(self, castObject: Cast, exception: Exception) -> None:
    """Initialize the CastException object."""
    self.cast = castObject
    self.value = castObject.__inner_value__
    self.target = castObject._getTargetType()
    self.why = exception
